import {MeterbusLibExceptions} from './exceptions'
import {MeterbusLibFrames} from './simpleframes'
import {MeterbusLibUtils} from './utils'
import {MeterbusLibCodeTables} from './codetables'

export namespace MeterbusLibLongFrame {
    export class LongFrame extends MeterbusLibFrames.ControlFrame {
        private _fixedDataHeader : FixedDataHeader
        private _variableDataBlocks : DataBlock[] = []

        get fixedDataHeader() : FixedDataHeader { return this._fixedDataHeader }

        constructor(telegram : number[]) {
            super(telegram)
        }

        parse2() {
            super.parse2() // control frame parse2 method

            let consumed : number = 7 // 68 L L 68 C A CI
            this._fixedDataHeader = new FixedDataHeader(this._telegram.slice(consumed))
            this._fixedDataHeader.parse()
            
            consumed += this._fixedDataHeader.consumed
            let count : number = 0
            while (true) {
                let die : DataBlock = 
                    new DataBlock(this._telegram.slice(consumed))
                die.parse()
                this._variableDataBlocks.push(die)
                consumed += die.consumed
                count++
                if ((consumed) >= (this._telegram.length - 2)) {
                    break                
                }
            }
        }
    }

    export class FixedDataHeader {
        private _data : number[]
        private __consumed : number = 0
        private _identNo : number
        private _manufacturer : string
        private _version : number
        private _medium : string
        private _accessNo : number
        private _status : number
        private _signature : number[]


        get identNo() : number { return this._identNo }
        get manufacturer() : string { return this._manufacturer }
        get version() : number { return this._version }
        get medium() : string { return this._medium }
        get accessNo() : number { return this._accessNo }
        get status() : number { return this._status }
        get signature() : number[] { return this._signature }

        constructor(data : number[]) {
            this._data = data
        }

        get consumed() : number {
            return this.__consumed
        }

        parse() : void {
            this._identNo = MeterbusLibUtils.bcd(this._data.slice(0, 4))
            this.__consumed += 4
            this._manufacturer = MeterbusLibUtils.manufCode(this._data.slice(4, 6))
            this.__consumed += 2
            this._version = this._data[6]
            this.__consumed += 1
            this._medium = MeterbusLibCodeTables.MEDIUM_CODE[this._data[7]]
            this.__consumed += 1
            this._accessNo = this._data[8]
            this.__consumed += 1
            this._status = this._data[9]
            this.__consumed += 1
            this._signature = this._data.slice(10, 12)
            this.__consumed += 2
        }

        toJSON() : any {
            return MeterbusLibUtils.jsonPrepaper(this, ["_data"])
        }
    }

    export abstract class ADataBlock {
        protected __consumed : number = 0

        get consumed() : number {
            return this.__consumed
        }

        abstract parse() : void

        toJSON() : any {
            return MeterbusLibUtils.jsonPrepaper(this, [])
        }
    }

    export class DataBlock extends ADataBlock {
        private _content : VariableDataBlock | ManufacturerSpecificData
        private _hex : string = ""
        private __data : number[]

        constructor(data : number[]) {
            super()
            this.__data = data
            if ((data[0] & 0x0f) == 0x0f) {
                this._content = new ManufacturerSpecificData(data)
            } else {
                this._content = new VariableDataBlock(data)
            }
        }

        get consumed() : number {
            return this._content.consumed
        }

        parse() : void {
            this._content.parse()

            for (let d of this.__data.slice(0, this.consumed)) {
                this._hex += "" + d.toString(16) + " "
            }

        }

        toJSON() : any {
            return MeterbusLibUtils.jsonPrepaper(this, [])
        }
    }

    export class DIF {
        raw : number
        function : number
        coding : number
        lsb : number
        extension : number

        toJSON() : any {
            let dup = MeterbusLibUtils.clone(this)
            dup.function = MeterbusLibCodeTables.DIF_FUNCTION_FIELD[dup.function]
            dup.coding = MeterbusLibCodeTables.DIF_CODING_FIELD[dup.coding].name
            return dup
        }
    }

    export class DIFE {
        raw : number
        extension : number
    }

    export class VIF {
        raw : number
        unitAndMultiplier : number
        extension : number
    }

    export class VIFE {
        raw : number
        extension : number
    }

    export class VariableDataBlock extends ADataBlock {
        private _type : string = "VDB"
        private __indata : number[]
        private _dif : DIF
        private _dife : DIFE[] = []
        private _vif : VIF
        private _vife : VIFE[] = []
        private _data : number[]
        private _value : any

        constructor(data : number[]) {
            super()
            this.__indata = data
        }

        parse() : void {
            this._dif = new DIF()
            this._dif.raw = this.__indata[0]
            this.__consumed++
            this._dif.function = (this._dif.raw & 0x30) >> 4
            this._dif.coding = this._dif.raw & 0x0f
            this._dif.lsb = (this._dif.raw & 0x40) >> 6
            this._dif.extension = (this._dif.raw & 0x80) >> 7

            if (this._dif.extension == 1) {
                while (true) {
                    let dife = new DIFE()
                    dife.raw = this.__indata[this.__consumed]
                    this.__consumed++
                    dife.extension = (dife.raw & 0x80) >> 7
                    this._dife.push(dife)
                    if (dife.extension == 0) {
                        break
                    }
                }
            }

            this._vif = new VIF()
            this._vif.raw = this.__indata[this.__consumed]
            this.__consumed++
            this._vif.unitAndMultiplier = this._vif.raw & 0x7f
            this._vif.extension = (this._vif.raw & 0x80) >> 7

            if (this._vif.extension == 1) {
                while (true) {
                    let vife = new VIFE()
                    vife.raw = this.__indata[this.__consumed]
                    this.__consumed++
                    vife.extension = (vife.raw & 0x80) >> 7
                    this._vife.push(vife)
                    if (vife.extension == 0) {
                        break
                    }
                }
            }
            
            let dcf : MeterbusLibCodeTables.DCFt = MeterbusLibCodeTables.DIF_CODING_FIELD[this._dif.coding]
            let actualLength = dcf.length
            if (dcf.length == -1) {
                actualLength = this.__indata[this.__consumed]
                this.__consumed++
            }
            this._data = this.__indata.slice(this.__consumed, 
                this.__consumed + actualLength)
            this.__consumed += actualLength

            this._value = dcf.encoder(this._data)
        }
    }

    export class ManufacturerSpecificData extends ADataBlock {
        private _type : string = "MSD"
        private __indata : number[]

        constructor(data : number[]) {
            super()
            this.__indata = data
        }

        parse() : void {
            
        }
    }

}